using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Lucene.Net.Search;
using Lucene.Net.Documents;
using System.Text.RegularExpressions;

namespace ClassificationApplication.FeatureExtractor
{
    /**<summary>Functionallity for extracting keywords and phrases from documents</summary>*/
    public class Extractor
    {
        /**<summary>ArrayList of documents to extract phrases from</summary>*/
        private ArrayList extractorDocuments;
        
        /**<summary>Index of phrases in the set of documents</summary>*/
        private Dictionary<Regex, ExtractorIndex> extractorIndices;
        //private ExtractorIndex index;

        /**<summary>List of resulting keyphrases</summary>*/
        private Tagger.Tagging.POSTagger tagger;
        //private ArrayList keyphrases;
        
        /**Constructor*/
        public Extractor(Hits retrievedDocuments)
        {
            tagger = new Tagger.Tagging.NorwegianPOSTagger(Tagger.Tagging.NorwegianPOSTagger.Lemmatize, true);
            extractorIndices = new Dictionary<Regex, ExtractorIndex>();
            if (retrievedDocuments.Length() > 0)
            {
                //Console.WriteLine("Extractor");
                for (int i = 0; i < retrievedDocuments.Length(); i++)
                {
                    Document doc = retrievedDocuments.Doc(i);
                    int docID = retrievedDocuments.Id(i);
                    //Console.WriteLine("Document id: " + docID);
                }
                //index = new ExtractorIndex();
                foreach (Regex regex in Globals.PosRegex)
                {
                    extractorIndices[regex] = new ExtractorIndex();
                }

                extractorDocuments = CreateExtractorDocuments(retrievedDocuments);
                
                //keyphrases = new ArrayList();
                foreach (ExtractorDocument exDoc in extractorDocuments)
                {
                    foreach (Regex regex in Globals.PosRegex)
                    {
                        //keyphrases.Add(
                        ExtractPhrases(regex, exDoc);
                    }
                    //);
                }
                //ArrayList combinedList = CombineResults(keyphrases);
            }
        }

        /**<summary>Collects phrases from all in one arraylist and sort it by weights</summary>
         * <param name="keyphrases">ArrayList containing an arraylist of phrases per document</param>
         * <returns>ArrayList of phrases in all documents sorted by weight</return>
         */
        private ArrayList CombineResults(ArrayList keyphrases)
        {
            ArrayList results = new ArrayList();
            foreach (ArrayList list in keyphrases) 
            {
                foreach (Keyphrase phrase in list) 
                {
                    results.Add(phrase);
                }
            }
            results.Sort();
            return results;
        }

        /**<summary>Transforms lucene-documents to ExtractorDocument objects</summary>
         * <param name="retrievedDocuments">Hits object containing lucene-document objects</param>
         * <returns>ArrayList of ExtractorDocument objects, one per lucene document in Hits-object</returns>
         */
        private ArrayList CreateExtractorDocuments(Hits retrievedDocuments) 
        {
            if (retrievedDocuments.Length() == 0) 
            {
                Console.WriteLine("No documents matches query");
                return null;
            }
            ArrayList exDocs = new ArrayList();
            for (int i = 0; i < retrievedDocuments.Length(); i++) 
            {
                Document doc = retrievedDocuments.Doc(i);
                int docID = retrievedDocuments.Id(i);
                exDocs.Add(new ExtractorDocument(doc, docID, extractorIndices,tagger));
            }
            return exDocs;
        }

        /** <summary>Calculates the TF-IDF score of the phrases contained in a document</summary>
         *  <param name="document">The document</param>
         *  <returns>Array list of weighted Keyphrase-objects</returns>
         */
        private List<Keyphrase> ExtractPhrases(Regex regex, ExtractorDocument document)
        {
            string[] phrases = document.GetPhrases(regex);
            int documentID = document.GetDocumentID();
            List<Keyphrase> weightedKeyphrases = new List<Keyphrase>(phrases.Length);
            foreach(String currentPhrase in phrases)
            {
                List<ExtractorWord> phraseOccurrences = document.GetOccurrences(regex,currentPhrase);
                double weight = CalculatePhraseWeight(regex, currentPhrase, phraseOccurrences, document);
                extractorIndices[regex].SetWeight(currentPhrase, documentID,weight); 

                //do we need the rest here?
                Keyphrase phrase = new Keyphrase(currentPhrase, weight, phraseOccurrences.Count);
                weightedKeyphrases.Add(phrase);
                //originalt blei den mest frekvente opprinnelege variant (utan stemming/lemmatisering) av frasen brukt for  representere frasen her
            }
            weightedKeyphrases.Sort();
            weightedKeyphrases.Reverse();
            return weightedKeyphrases;
        }

        /**<summary>Calculates the weight of a phrase using a TF-IDF scheme</summary>*/
        private double CalculatePhraseWeight(Regex regex, String currentPhrase, List<ExtractorWord> phraseOccurrences, ExtractorDocument document)
        {
            //Calculate weight
            double weight;
            //The weight of the given phrase (SUM(count*sourceWeight))
            double phraseOccurenceWeight = 0;
            foreach (ExtractorWord ew in phraseOccurrences)
            {
                phraseOccurenceWeight += ew.Weight;
            }
            if (Globals.doTFIDFWeighting)
            {
                //TFIDFweighting, 
                double tf = phraseOccurenceWeight / (double)document.GetNofWords(regex);
                double idf = -(double)((double)Math.Log(((double)extractorIndices[regex].GetNofDocsContainingWord(currentPhrase)) / ((double)extractorDocuments.Count))) / ((double)Math.Log(2.0d));
                //double idf = Math.Log((double)extractorDocuments.Count / (double)extractorIndices[regex].GetNofDocsContainingWord(currentPhrase));
                weight = tf * idf;
                if (weight == 0)
                {

                }
                //Console.WriteLine(extractorDocuments.Count);
            }
            else
            {
                double tf = phraseOccurenceWeight / (double)document.GetNofWords(regex);
                double df = Math.Log(1 + extractorIndices[regex].GetNofDocsContainingWord(currentPhrase));
                weight = tf * df;
            }

            //Boost multiword-phrases
            if (Globals.doMultiWordPhraseBoost)
            {
                if (currentPhrase.Split(' ').Length > 1)
                {
                    weight *= Globals.MultiWordPhraseBoost(regex);
                }
            }

            //Boost wordlength
            if (Globals.doWordLengthBoost)
            {
                double wordLength = (double)currentPhrase.Length;
                weight *= 1.0d + (wordLength / Globals.wordLengthBoostFactor);
            }
            return weight;
        }

        public ExtractorIndex GetIndex(Regex regex) 
        {
            if (extractorIndices.ContainsKey(regex))
                return extractorIndices[regex];
            else
                return new ExtractorIndex();
        }
    }
}
