using System;
using System.Collections.Generic;
using System.Text;
using Lucene.Net.Documents;
using System.Collections;
using Tagger.Tagging;
using System.Text.RegularExpressions;

namespace ClassificationApplication.FeatureExtractor
{
    /**<summary>Encapsulates a Lucene-Document-Object. Suitable for extracting phrases.</summary>*/
    class ExtractorDocument 
    {
        /**<summary>The encapsulated luceneDocument</summary>*/
        private Document luceneDocument;
        
        /**<summary>Hashtable for fast lookup of occurences</summary>*/
        //private Hashtable occurrences;

        private Dictionary<Regex,Hashtable> regexOccurrences;

        /**<summary>The number of distinct words, i.e. phrases in this document</summary>*/
        private Dictionary<Regex, int> nofWords;
        //private int nofWords;

        /**<summary>The id of the document</summary>*/
        private int documentID;

        private POSTagger tagger;
        /**<summary>Constructor</summary>
         * <param name="documentID">The id of the Lucene-document, also used as id here.</param>
         * <param name="index">The index of the belonging resultset.</param>
         * <param name="luceneDocument">The Lucene-Document to build this ExtractorDocument from.</param>
         */
        public ExtractorDocument(Document luceneDocument, int documentID, Dictionary<Regex, ExtractorIndex> index, POSTagger tagger)
        {
            this.tagger = tagger;
            regexOccurrences = new Dictionary<Regex, Hashtable>();
            nofWords = new Dictionary<Regex, int>();
            this.luceneDocument = luceneDocument;
            this.documentID = documentID;
            //this.occurrences = new Hashtable();
            //this.nofWords = 0;
            
            List<ExtractorWord> alLists = ExtractPatterns();
            
            AddPhrasesToLocalIndex(alLists);

            foreach (Regex regex in Globals.PosRegex)
            {
                if (regexOccurrences.ContainsKey(regex))
                {
                    index[regex].AddOccurrences(documentID, regexOccurrences[regex],regex);
                }
                //index.AddOccurrences(documentID, alLists);
            }

        }

        public void PrintResult(ArrayList list) 
        {
            ArrayList nounPhrases = (ArrayList)list[0];
            foreach (ExtractorWord noun in nounPhrases) 
            {
                Console.WriteLine("Phrase: " + noun.phrase);
            }
            ArrayList verbs = (ArrayList)list[1];
            foreach (ExtractorWord verb in verbs) 
            { 
                Console.WriteLine("Verb: " +verb.phrase);
            }
        }

        private List<ExtractorWord> ExtractPatterns()
        {
            string[] fieldText = new string[Globals.TaggedFields.Length];
            int i = 0;
            foreach (string field in Globals.TaggedFields)
            {
                fieldText[i] = luceneDocument.GetField(field).StringValue();
                //fieldText[i] = luceneDocument.GetField(field).StringValue();
                i++;
            }
            string urlDomain = luceneDocument.GetField("url").StringValue();
            urlDomain = urlDomain.Substring(urlDomain.IndexOf("http://") + 7);
            urlDomain = urlDomain.Substring(0, urlDomain.IndexOf("/"));
            //ArrayList phrases = new ArrayList();
            List<ExtractorWord> phrases = new List<ExtractorWord>();

            foreach (string field in fieldText)
            {

                //string[] taggedString = tagger.Tag(field.StringValue());
                //foreach (string taggedSentence in taggedString)
                //{
                string[] tokenizedField = field.Split(' ');//taggedSentence.Split(' ');
                string[] tags = new string[tokenizedField.Length];
                string[] words = new string[tokenizedField.Length];

                i = 0;
                string[] parts;
                foreach (string token in tokenizedField)
                {
                    parts = token.Split('/');
                    words[i] = parts[0];
                    tags[i] = parts[1];
                    i++;
                }
                string tagString = tags[0];
                for (int j = 1; j < tags.Length; j++)
                {
                    tagString = string.Concat(tagString, " ", tags[j]);
                }
                int[] tagPositions = new int[tagString.Length + 1];
                int tmpPos = 0;
                //int currentPos = 0;
                tagPositions[0] = 1;
                tagPositions[0 + tags[0].Length - 1] = 1;
                for (int j = 1; j < tags.Length; j++)
                {
                    tmpPos += tags[j - 1].Length + 1;
                    tagPositions[tmpPos] = j + 1;
                    tagPositions[tmpPos + tags[j].Length - 1] = j + 1;
                }

                foreach (Regex regex in Globals.PosRegex)
                {

                    Match m = regex.Match(tagString);
                    //Match nextMatch;

                    while (m.Success)
                    {
                        try
                        {
                            int startIndex = tagPositions[m.Index] - 1;//m.Index;
                            int endIndex = tagPositions[m.Index + m.Length - 1] - 1;//m.Index + m.Length;
                            string phrase = words[startIndex];
                            string pos = tags[startIndex];

                            for (int j = startIndex + 1; j <= endIndex; j++)
                            {
                                phrase = string.Concat(phrase, " ", words[j]);
                                pos = string.Concat(pos, " ", tags[j]);
                            }
                            phrases.Add(new ExtractorWord(phrase, regex, urlDomain));

                            m = m.NextMatch();
                        }
                        catch (IndexOutOfRangeException afsd) { m = m.NextMatch(); }

                    }

                    //}
                }
            }
            return phrases;
        }
        /*
        public static void Main()
        {
            string text = "One car red car blue car";
            string pat = @"(\w+)\s+(car)";
            // Compile the regular expression.
            Regex r = new Regex(pat, RegexOptions.IgnoreCase);
            // Match the regular expression pattern against a text string.
            Match m = r.Match(text);
            int matchCount = 0;
            while (m.Success)
            {
                Console.WriteLine("Match" + (++matchCount));
                for (int i = 1; i <= 3; i++)
                {
                    Group g = m.Groups[i];
                    Console.WriteLine("Group" + i + "='" + g + "'");
                    CaptureCollection cc = g.Captures;
                    for (int j = 0; j < cc.Count; j++)
                    {
                        Capture c = cc[j];
                        System.Console.WriteLine("Capture" + j + "='" + c + "', Position=" + c.Index);
                    }
                }
                m = m.NextMatch();
            }
        }
         */ 
        /** <summary>Builds the hashtable of occurrences of this document</summary>
         *  <param name="phrases">The phrases to add to the hashtable</phrases>
         */
        private void AddPhrasesToLocalIndex(List<ExtractorWord> phrases)
        {
            foreach (ExtractorWord phrase in phrases)
            {
                if (!regexOccurrences.ContainsKey(phrase.Regex))
                {
                    regexOccurrences[phrase.Regex] = new Hashtable();
                }
                Hashtable occurrences = regexOccurrences[phrase.Regex];
                //all words are kept in arraylist, this is most likely unecessary (originally to later find the most used original variant of the word), enough to keep the count...
                if (!occurrences.ContainsKey(phrase.phrase))
                {
                    //Doesn't exist, add new occurrence to this word
                    List<ExtractorWord> newList = new List<ExtractorWord>();
                    newList.Add(phrase);
                    occurrences.Add(phrase.phrase, newList);
                    //occurrences.Add(phrase.word, phrase);
                }
                else
                {
                    //Add an occurrence
                    List<ExtractorWord> list = (List<ExtractorWord>)occurrences[phrase.phrase];
                    list.Add(phrase);
                    
                    //((ExtractorWord)occurrences[phrase.word]).IncrementCount();
                }
                if (!nofWords.ContainsKey(phrase.Regex))
                    nofWords[phrase.Regex] = 1;
                else
                {
                    int tmpCount = nofWords[phrase.Regex];
                    nofWords.Remove(phrase.Regex);
                    nofWords[phrase.Regex] = tmpCount++;
                }
            }
        }
        
        /**<summary>Getter for the phrases of this document</summary>
         * <returns>A String[] of phrases, one entry per phrase</returns>
         */
        public String[] GetPhrases(Regex regex)
        {
            if (regexOccurrences.ContainsKey(regex))
            {

                Hashtable occurrences = regexOccurrences[regex];
                IDictionaryEnumerator myEnumerator = occurrences.GetEnumerator();
                string[] keys = new String[occurrences.Count];
                int i = 0;
                while (myEnumerator.MoveNext())
                {
                    keys[i] = (String)myEnumerator.Key;
                    i++;
                }
                return keys;
            }
            return new string[] { };
        }

        /**<summary>Gets all occurrences of a given phrase</summary>
         * <param name="phraseKey">The lookup phrase for return of occurrences</param> 
         * <returns>ArrayList of occurrences of the given phrase</returns>
         */
        public List<ExtractorWord> GetOccurrences(Regex regex,String phraseKey)
        {
            Hashtable occurrences = regexOccurrences[regex];
            return (List<ExtractorWord>)occurrences[phraseKey];
        }

        /**<summary>Gets the number of distinct words, i.e. phrases in the document</summary>
         * <returns>The number of distinct words in the document</returns>
         */
        public int GetNofWords(Regex regex) 
        {
            return nofWords[regex];
        }

        public int GetDocumentID() 
        {
            return documentID;
        }
    }
}
