using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;

namespace ClassificationApplication.FeatureExtractor
{
    public class ExtractorIndex : Index
    {

        /**<summary>A hashtable of words in collection as keys for fast lookup. Each entry is a new Hashtable of documents, each with entry holding the count and weight of the word in the document</summary>*/
        private Hashtable occurrences = new Hashtable();
        private List<string> words = new List<string>();

        /**<summary>Needed to compute cosine similarity</summary>*/
        private double norm = 1;

        public void AddOccurrences(int documentID, Hashtable words, Regex regex)
        {
            foreach (string key in words.Keys)
            {
                foreach (ExtractorWord eWord in (List<ExtractorWord>)words[key])
                {
                    AddOccurence(documentID, eWord, regex);
                }
            }
        }

        private void AddOccurence(int documentID, ExtractorWord word, Regex regex)
        {
            //check if word exsists in index
            if (!occurrences.ContainsKey(word.phrase))
            {
                //add the new word
                occurrences.Add(word.phrase, new Hashtable());
            }

            //do lookup
            Hashtable docs = (Hashtable)occurrences[word.phrase];

            //check if document already registered with the current word
            if (!docs.ContainsKey(documentID))
            {
                //add the new document 
                //ArrayList newList = new ArrayList();
                //newList.Add(word);
                //docs.Add(documentID, newList);
                //Create own ExtractorHashEntry?
                ExtractorIndexEntry entry = new ExtractorIndexEntry(word.phrase);
                docs.Add(documentID, entry);
            }
            else
            {
                //document already exists in the index, add the word in its list 
                //ArrayList list = (ArrayList)docs[documentID];
                //list.Add(word);
                ((ExtractorIndexEntry)docs[documentID]).IncrementCount();
            }
        }

        public double GetNofDocsContainingWord(String word)
        {
            if (!occurrences.ContainsKey(word))
            {
                return 0;
            }
            Hashtable docs = (Hashtable)occurrences[word];
            return (double)docs.Count;
        }

        /**<summary>Sums the weights of the word in all documents</summary>
         * <param name="word">The word to sum the weights of</param>
         * <returns>The summed weight of the word in the index</returns>
         */
        public double GetSummedWeight(String word)
        {
            if (!occurrences.ContainsKey(word))
            {
                return 0.0d;
            }
            else
            {
                double summedWeight = 0.0d;
                Hashtable docs = (Hashtable)occurrences[word];
                IEnumerator enumerator = docs.GetEnumerator();
                {
                    while (enumerator.MoveNext())
                    {
                        ExtractorIndexEntry entry = (ExtractorIndexEntry)enumerator.Current;
                        summedWeight += entry.GetWeight();
                    }
                }
                return summedWeight;
            }
        }

        /**<summary>Sums and normalizes the weights of the word in all documents</summary>
         * <param name="word">The word to sum the weights of</param>
         * <returns>The normalized weight of the word in the index</returns>
         */
        public double GetNormalizedSummedWeight(String word)
        {
            if (!occurrences.ContainsKey(word))
            {
                return 0.0d;
            }
            else
            {
                double summedWeight = 0.0d;
                Hashtable docs = (Hashtable)occurrences[word];

                foreach (System.Collections.DictionaryEntry dictionaryEntry in docs)
                {
                    ExtractorIndexEntry indexEntry = (ExtractorIndexEntry)dictionaryEntry.Value;
                    summedWeight += indexEntry.GetWeight();
                }

                /*
                IEnumerator enumerator = docs.GetEnumerator();
                {
                    while ( enumerator.MoveNext())
                    {
                        ExtractorIndexEntry entry = (ExtractorIndexEntry)enumerator.Current;
                        summedWeight += entry.GetWeight();
                    }
                }
                 */
                return summedWeight / docs.Count;
            }
        }

        public void SetWeight(string key, int documentID, double weight)
        {
            ((ExtractorIndexEntry)((Hashtable)(occurrences[key]))[documentID]).SetWeight(weight);

            // increase the norm
            norm += weight * weight;
        }

        public double GetWeight(string word)
        {
            double weight;
            if (Globals.doNormalizeWeights)
            {
                weight = GetNormalizedSummedWeight(word);
            }

            else
            {
                weight = GetSummedWeight(word);
            }
            return weight;
        }

        public ICollection GetKeys()
        {
            return occurrences.Keys;
        }

        public bool ContainsKey(string word)
        {
            return occurrences.ContainsKey(word);
        }

        public double GetNorm()
        {
            return norm;
        }

        public double Compare(Index index2)
        {
            using (System.IO.StreamWriter wr = new System.IO.StreamWriter("c:\\features.txt", false, System.Text.Encoding.Default))
            {
                double weightProd = 0.0;
                foreach (string key in GetKeys())
                {
                    if (index2.ContainsKey(key))
                    {
                        weightProd += GetWeight(key) * index2.GetWeight(key);
                        wr.WriteLine(key+"\t"+GetWeight(key)*index2.GetWeight(key));
                    }
                }
                double norm = Math.Sqrt(GetNorm()) * Math.Sqrt(index2.GetNorm());
                double compareValue = weightProd / norm;
                return compareValue;
            }
        }

        public SortedList<double, string> GetSortedIndex()
        {
            ICollection keys = GetKeys();
            SortedList<double, string> sortedList = new SortedList<double, string>();
            double weight;
            foreach (string key in keys)
            {
                weight = GetWeight(key);
                while (sortedList.ContainsKey(weight))
                {
                    weight += .0001d;
                }
                if (weight > 0.1)
                {
                    sortedList[weight] = key;
                }
            }
            return sortedList;
        }
    }
}
